class Type:
    NOOP, BACKGROUND, FOREGROUND, MODIFIER, PAYLOAD = range(5)


class Node:
    def traverse(self, node):
        if node.type == Type.FOREGROUND and self.type == Type.FOREGROUND:
            self.type = Type.BACKGROUND
            return

        for each in node.children:
            self.traverse(each)

    def __init__(self, type, value, *children, **kwargs):
        bg = kwargs.get('bg', False)

        self.type = Type.BACKGROUND if bg else type
        self.value = value
        self.children = []

        for each in children:
            if isinstance(each, str):
                self.children.append(Node(Type.PAYLOAD, each))

            elif isinstance(each, Node):
                self.traverse(each)
                self.children.append(each)

    def __add__(self, other):
        root = Node(Type.NOOP, 0, self, other)
        return root

    def __repr__(self):
        names = {
            Type.NOOP: 'noop',
            Type.BACKGROUND: 'background',
            Type.FOREGROUND: 'foreground',
            Type.MODIFIER: 'modifier',
            Type.PAYLOAD: 'payload',
        }

        return names[self.type]

    def __str__(self):
        return render(self)


class SGR:
    def __init__(self, *parameters):
        self.parameters = set(parameters)

    def __eq__(self, other):
        return self.parameters == other.parameters

    def __str__(self):
        parameters = map(str, self.parameters)
        return '\x1b[' + ';'.join(parameters) + 'm'

    def transition(self, other):
        difference = other.parameters - self.parameters
        return self.__class__(*difference)


def walker(node, paths=None):
    root = []
    fg, bg, mod = False, False, False

    def walker_(node, paths):
        if node.type != Type.PAYLOAD:
            for each in node.children:
                walker_(each, paths + [node])
        else:
            root.append(paths + [node])

    walker_(node, [])
    return root


def render(node):
    result = ''
    current = SGR()

    for path in walker(node):
        payloads = []

        parameters = []
        for node in reversed(path):
            if node.type == Type.PAYLOAD:
                payloads.append(node.value)
                continue

            if node.type == Type.FOREGROUND:
                parameters.append(30 + node.value)

            if node.type == Type.BACKGROUND:
                parameters.append(40 + node.value)
                break

            if node.type == Type.MODIFIER:
                parameters.append(node.value)

        future = SGR(*parameters)

        if not current == future:
            transition = current.transition(future)
            result += str(transition)
            current = future

        result += ''.join(payloads)

    return '\x1b[0m' + result + '\x1b[0m'
